extern crate reactive_net;

mod __authentic_execution;
pub mod __run;

#[allow(unused_imports)] use __authentic_execution::authentic_execution;
#[allow(unused_imports)] use __authentic_execution::authentic_execution::{MODULE_NAME, success, failure, handle_output, handle_request, Error};
#[allow(unused_imports)] use reactive_net::{ResultCode, ResultMessage};
